/*
 * devtree.c -- represent the tree of device nodes, both found and needed
 *
 * Copyright 2007-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Peter Jones <pjones@redhat.com>
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <nash.h>
#include "list.h"
#include "util.h"
#include "devtree.h"
#include "lib.h"
#include "dm.h"

extern int nash_bdev_vitals_get_type(struct nash_bdev_vitals *);

extern void nash_dev_node_free_ptr(nashContext *nc, 
    struct nash_dev_node **nodep);
static struct nash_dev_node *nash_dev_node_from_blkent(nashContext *nc,
    struct blkent *blkent);
static struct nash_dev_node *nash_dev_node_from_bdev(nashContext *nc,
    struct nash_block_dev *bdev);
static void nash_dev_node_emancipate_from_parents(nashContext *nc,
    struct nash_dev_node *child);
static void nash_dev_node_surrender_children(nashContext *nc, 
    struct nash_dev_node *adult);
static int nash_dev_node_viable_parent(nashContext *nc,
        struct nash_dev_node *adult,
        struct nash_dev_node *child);

/*
 * First: the structure that goes in nashContext and its accessors
 */

struct nash_dev_tree; /* actually it's in the header now */

typedef int (*nash_dev_tree_visitor)(nashContext *nc,
    struct nash_dev_node *node,
    void *priv);

int nash_dev_node_visit_children_helper(nashContext *nc,
    struct nash_dev_node *adult, int visit_adult, nash_dev_tree_visitor visitor,
    struct nash_list *to_visit, void *priv)
{
    struct nash_list_iter *i;
    struct nash_dev_node *child;
    int rc;
    int pos;

    rc = 0;

    pos = nash_list_search(to_visit, adult);
    if (visit_adult && pos >= 0) {
        nash_list_remove(to_visit, pos);
        rc += visitor(nc, adult, priv);
    }

    i = NULL;
    while ((i = nash_list_iter_next(adult->children, i))) {
        child = nash_list_iter_data(i);
        if ((pos = nash_list_search(to_visit, child)) >= 0) {
            rc += nash_dev_node_visit_children_helper(nc, child, 1, visitor,
                to_visit, priv);
        }
    }

    return rc;
}

int nash_dev_node_visit_children(nashContext *nc, struct nash_dev_node *adult, 
    int visit_adult, nash_dev_tree_visitor visitor, void *priv)
{
    struct nash_dev_tree *tree = adult->tree;
    struct nash_list *to_visit;
    int errnum;
    int rc;

    if (!(to_visit = nash_ptr_list_new()))
        return -1;

    struct nash_list_iter *i;
    i = NULL;
    while ((i = nash_list_iter_next(tree->devs, i))) {
        struct nash_dev_node *node = nash_list_iter_data(i);
        nash_list_add(to_visit, node);
    }

    rc = nash_dev_node_visit_children_helper(nc, adult, visit_adult, visitor,
        to_visit, priv);

    save_errno(errnum, nash_list_destroy(to_visit));
    return rc;
}

void nash_dev_tree_delete_node(nashContext *nc, struct nash_dev_node *node)
{
    struct nash_dev_tree *tree = node->tree;
    int pos;

    nash_dev_node_emancipate_from_parents(nc, node);
    nash_dev_node_surrender_children(nc, node);
    if (node->type == DEV_TYPE_FS) {
        if ((pos = nash_list_search(tree->fs_devs, node)) >= 0)
            nash_list_remove(tree->fs_devs, pos);
    }

    if ((pos = nash_list_search(tree->devs, node)) >= 0)
        nash_list_remove(tree->devs, pos);
}

static int attach_parents_visitor(nashContext *nc, struct nash_dev_node *adult, 
    void *childp)
{
    struct nash_dev_node *child = childp;

    if (nash_list_in(child->parents, adult))
        return 0;
    if (nash_dev_node_viable_parent(nc, adult, child)) {
        nash_list_add(adult->children, child);
        nash_list_add(child->parents, adult);
        nash_dev_tree_check_node_complete(nc, child);
        return 1;
    }
    return 0;
}

static int attach_parents(nashContext *nc, struct nash_dev_tree *tree)
{
    struct nash_list_iter *adults, *children;
    struct nash_dev_node *adult, *child;
    int ret, changed, rc;

    ret = 0;
    do {
        changed = ret;

        children = NULL;
        while (changed == ret &&
                (children = nash_list_iter_next(tree->devs, children))) {
            child = nash_list_iter_data(children);
            adults = NULL;
            while (changed == ret &&
                    (adults = nash_list_iter_next(tree->devs, adults))) {
                adult = nash_list_iter_data(adults);
                rc = nash_dev_node_visit_children(nc, adult, 1, 
                    attach_parents_visitor, child);
                ret += rc;
                if (changed != ret)
                    break;
            }
            if (adults)
                nash_list_iter_end(adults);
        }
        if (children)
            nash_list_iter_end(children);
    } while (changed != ret);
    return ret;
}

#ifdef DO_PRINT_DEV_TREE
extern int print_vitals(struct nash_bdev_vitals *vitals);
static int print_dev_tree_visitor(nashContext *nc, struct nash_dev_node *child, 
    void *priv)
{
    struct nash_list_iter *iter;

    fprintf(stderr, "node: %s ", child->name);
    if (child->mate)
        fprintf(stderr, "is mated to %s\n", child->mate->name);
    else
        fprintf(stderr, "is not mated\n");

    if (nash_list_len(child->vitals) > 0) {
        int printed = 0;
        iter = NULL;
        while ((iter = nash_list_iter_next(child->vitals, iter))) {
            struct nash_bdev_vitals *vitals = nash_list_iter_data(iter);
            if (!printed)
                printed = print_vitals(vitals);
        }
    } else {
        fprintf(stderr, "  no vitals for %s\n", child->name);
    }
    if (child->type != DEV_TYPE_DISK && child->type != DEV_TYPE_DM_ERROR) {
        if (nash_list_len(child->parent_names) > 0) {
            fprintf(stderr, "  parent names: ");
            iter = NULL;
            while ((iter = nash_list_iter_next(child->parent_names, iter))) {
                char *name = nash_list_iter_data(iter);
                fprintf(stderr, "%s ", name);
            }
            fprintf(stderr, "\n");
        }
        if (nash_list_len(child->parents) > 0) {
            fprintf(stderr, "  parents: ");
            iter = NULL;
            while ((iter = nash_list_iter_next(child->parents, iter))) {
                struct nash_dev_node *node = nash_list_iter_data(iter);
                fprintf(stderr, "%s ", node->name);
            }
            fprintf(stderr, "\n");
        }
    }
    if (nash_list_len(child->child_names) > 0) {
        fprintf(stderr,"  child names: ");
        iter = NULL;
        while ((iter = nash_list_iter_next(child->child_names, iter))) {
            char *name = nash_list_iter_data(iter);
            fprintf(stderr,"%s ", name);
        }
        fprintf(stderr,"\n");
    }
    if (nash_list_len(child->children) > 0) {
        fprintf(stderr,"  children: ");
        iter = NULL;
        while ((iter = nash_list_iter_next(child->children, iter))) {
            struct nash_dev_node *node = nash_list_iter_data(iter);
            fprintf(stderr,"%s ", node->name);
        }
        fprintf(stderr,"\n");
    }

    return 0;
}

void print_dev_tree(nashContext *nc, int blktab)
{
    struct nash_dev_node *root = blktab ? nc->blktab->root : nc->devices->root;
    fprintf(stderr,"printing %s tree\n", blktab ? "blktab": "device");
    nash_dev_node_visit_children(nc, root, 0, print_dev_tree_visitor, NULL);
}
#else
void print_dev_tree(nashContext *nc, int blktab) {
    return;
}
#endif

static int ok_type_combo(struct nash_dev_node *left,
        struct nash_dev_node *right)
{
    int type0 = left->type, type1 = right->type;
    if (type0 == DEV_TYPE_TREE || type1 == DEV_TYPE_TREE)
        return 0;

    if (type0 == DEV_TYPE_NONE || type1 == DEV_TYPE_NONE)
        return 0;

    switch (type0) {
        case DEV_TYPE_DISK:
        case DEV_TYPE_DM_ERROR:
            if (type1 == DEV_TYPE_DISK || type1 == DEV_TYPE_DM_ERROR)
                return 1;
            return 0;
        case DEV_TYPE_DM:
            switch (type1) {
                case DEV_TYPE_DM:
                case DEV_TYPE_DM_MPATH:
                case DEV_TYPE_LVM2_LV:
                case DEV_TYPE_DM_RAID:
                    return 1;
                case DEV_TYPE_PARTITION:
                    if (right->flags & DEV_FLAG_DM)
                        return 1;
            }
            return 0;
        case DEV_TYPE_DM_MPATH:
            if (type1 == DEV_TYPE_DM || type1 == DEV_TYPE_DM_MPATH)
                return 1;
            return 0;
        case DEV_TYPE_LVM2_LV:
            if (type1 == DEV_TYPE_DM || type1 == DEV_TYPE_LVM2_LV)
                return 1;
            return 0;
        case DEV_TYPE_DM_RAID:
            if (type1 == DEV_TYPE_DM || type1 == DEV_TYPE_DM_RAID)
                return 1;
            return 0;
        case DEV_TYPE_PARTITION:
            switch (type1) {
                case DEV_TYPE_DM:
                    if (left->flags & DEV_FLAG_DM)
                        return 1;
                    break;
                case DEV_TYPE_PARTITION:
                    if (left->flags & DEV_FLAG_DM && right->flags & DEV_FLAG_DM)
                        return 1;
                    if (!(left->flags & DEV_FLAG_DM)
                            && !(right->flags & DEV_FLAG_DM))
                        return 1;
            }
            return 0;
        case DEV_TYPE_LVM2_PV:
        case DEV_TYPE_LVM2_VG:
        case DEV_TYPE_MD_RAID:
        case DEV_TYPE_FS:
            if (left->type == right->type)
                return 1;
    }
    return 0;
}

static int tree_propogate_flags(nashContext *nc, struct nash_dev_node *child, 
    void *priv)
{
    struct nash_list_iter *iter;

    iter = NULL;
    while ((iter = nash_list_iter_next(child->parents, iter))) {
        struct nash_dev_node *parent = nash_list_iter_data(iter);

        if (parent->type == DEV_TYPE_TREE || parent->type == DEV_TYPE_NONE)
            continue;
        if (parent->flags & DEV_FLAG_ACCESS_OK) {
            child->flags |= DEV_FLAG_ACCESS_OK;
            child->flags &= ~DEV_FLAG_PROTECTED;
        }
    }
    return 0;
}

static int nash_dev_node_try_mate(nashContext *nc, struct nash_dev_node *dev,
        struct nash_dev_node *tab)
{
    struct nash_list_iter *vi0;
    int agree;
    int rc;

    if (tab->mate || dev->mate)
        return tab->mate == dev->mate;

    if (!ok_type_combo(tab, dev))
        return 0;

    vi0 = NULL;
    agree = 0;
#if 0
    fprintf(stderr, "%s checking vitals for %s and %s ...\n", __func__,
            dev->bdev->sysfs_path, tab->name);
#endif
    while ((vi0 = nash_list_iter_next(dev->vitals, vi0))) {
        struct nash_list_iter *vi1 = NULL;
        struct nash_bdev_vitals *v0 = nash_list_iter_data(vi0);

        while ((vi1 = nash_list_iter_next(tab->vitals, vi1))) {
            struct nash_bdev_vitals *v1 = nash_list_iter_data(vi1);
#if 0
            fprintf(stderr, "%d comparing vitals %s (%p) and %s (%p)\n",
                    __LINE__, v0->name, v0, v1->name, v1);
#endif
            if ((rc = nash_bdev_vitals_type_cmp(v0, v1))) {
#if 0
                fprintf(stderr, "vitals have different types (%d and %d)\n",
                        v0->type, v1->type);
#endif
                continue;
            }
            rc = nash_bdev_vitals_cmp(nc, v0, v1);
#if 0
            fprintf(stderr, "nash_bdev_vitals_cmp returned %d\n", rc);
#endif
            if (rc != 0) {
                agree = -1;
                nash_list_iter_end(vi1);
                break;
            } else if (agree >= 0) {
                agree++;
            }
        }
        if (agree < 0) {
            nash_list_iter_end(vi0);
            break;
        }
    }
#if 0
    fprintf(stderr, " agree was %d\n", agree);
#endif
    if (agree < 1)
        return 0;

    nashLogger(nc, NASH_NOTICE, "Found device %s as %s\n", tab->name, dev->bdev->sysfs_path);
    tab->mate = dev;
    dev->mate = tab;

    dev->flags |= tab->flags;
    if (dev->flags & DEV_FLAG_ACCESS_OK) {
        dev->flags &= ~DEV_FLAG_PROTECTED;
        if (dev->type == DEV_TYPE_DISK) {
            dev->flags |= DEV_FLAG_COMPLETE;
            tab->flags |= DEV_FLAG_COMPLETE;
        }
    }

    nash_dev_node_visit_children(nc, dev, 0, tree_propogate_flags, NULL);
    return 1;
}

/* XXX this algorithm sucks ass.  need to make these be hashed on their
 * vital signs instead. */
static int nash_dev_tree_group_nodes(nashContext *nc,
        struct nash_dev_node *node)
{
    struct nash_dev_tree *tree = nc->blktab;
    struct nash_list_iter *iter;
    int rc = 0;

    iter = NULL;
    while ((iter = nash_list_iter_next(tree->devs, iter))) {
        struct nash_dev_node *other = nash_list_iter_data(iter);
        if (nash_dev_node_try_mate(nc, node, other)) {
            rc = 1;
            nash_list_iter_end(iter);
            break;
        }
    }
    iter = NULL;
    while (rc && (iter = nash_list_iter_next(node->children, iter))) {
        struct nash_dev_node *child = nash_list_iter_data(iter);
        if (!(child->flags & DEV_FLAG_COMPLETE))
            nash_dev_tree_group_nodes(nc, child);
        nash_dev_tree_check_node_complete(nc, child);
    }

    return rc;
}

static int nash_dev_tree_add_node(nashContext *nc, int blktab,
        struct nash_dev_node *node)
{
    struct nash_dev_tree *tree = blktab ? nc->blktab : nc->devices;
    struct nash_dev_node *root = tree->root;

    /*
     * If it's a filesystem, we need to add it to the fs list, and then put
     *   it into the tree at as low of a level as is allowed.
     * 
     * If anything topology changed under a disk, we need to tell our caller.
     */
    nash_list_add(tree->devs, node);
    if (node->type == DEV_TYPE_DISK || node->type == DEV_TYPE_DM_ERROR) {
        nash_list_add(root->children, node);
        nash_dev_tree_check_node_complete(nc, node);
        node->flags |= DEV_FLAG_IN_TREE;
    } else {
        if (node->type == DEV_TYPE_FS)
            nash_list_add(tree->fs_devs, node);
    }
    return attach_parents(nc, tree);
}

int nash_dev_tree_add_blkent(nashContext *nc, struct blkent *blkent)
{
    struct nash_dev_node *node;
    int rc, i;

    if (!(node = nash_dev_node_from_blkent(nc, blkent)))
        return -1;

    if ((rc = nash_dev_tree_add_node(nc, 1, node)) >= 0) {
        if (nc->devices) {
            for (i = 0; (node = nash_list_get(nc->devices->devs, i)) ; i++) {
                if (node->type == DEV_TYPE_DISK && !node->mate)
                    nash_dev_tree_group_nodes(nc, node);
                nash_dev_tree_check_node_complete(nc, node);
            }
        }
    }
    return rc;
}

struct nash_dev_node *nash_dev_tree_process_bdev(nashContext *nc,
        struct nash_block_dev *bdev)
{
    struct nash_dev_node *node;
    int rc;
    int errnum;

#if 0
    fprintf(stderr, "%s processing bdev for %s\n", __func__, bdev->sysfs_path);
#endif
    switch (bdev->type) {
        case ADD:
            if (!strlimitcmp("/sys/block/ram", bdev->sysfs_path) ||
                    !strlimitcmp("/sys/block/loop", bdev->sysfs_path)) {
                nashBdevFreePtr(&bdev);
                return NULL;
            }

            if (!(node = nash_dev_node_from_bdev(nc, bdev))) {
#if 0
                fprintf(stderr, "%s could not make dev node from %s\n",
                        __func__, bdev->sysfs_path);
#endif
                return NULL;
            }

            if (node->type == DEV_TYPE_NONE) {
#if 0
                fprintf(stderr, "%s node type was none for %s\n",
                        __func__, bdev->sysfs_path);
#endif
                nash_dev_node_free_ptr(nc, &node);
                return NULL;
            }

            if (nc->blktab) {
                if (node->type == DEV_TYPE_DISK) {
#if 0
                    fprintf(stderr, "%s was type disk, grouping nodes\n", bdev->sysfs_path);
#endif
                    nash_dev_tree_group_nodes(nc, node);
                } else
#if 0
                    fprintf(stderr, "%s node type for %s was not disk, not grouping\n", __func__, bdev->sysfs_path);
#endif
                nash_dev_tree_check_node_complete(nc, node);
            }

            if ((rc = nash_dev_tree_add_node(nc, 0, node)) < 0) {
                save_errno(errnum, nash_dev_node_free_ptr(nc, &node));
                return NULL;
            }

            return node;
        case REMOVE:
            /* XXX FIXME */
            return NULL;
        case UNALLOCATED:
            return NULL;
    }
    /* never reached */
    return NULL;
}

static int
try_sysfs_dir(char *sysfs_path, struct nash_block_dev **dev)
{
    int ret;
    dev_t devno = 0;
    char *bang = NULL;
    struct nash_block_dev *tmp = NULL;

    nassert(sysfs_path && sysfs_path[0]);
    /* < 0 means other error
     *   0 means "no sysfs_path/dev file"
     * > 0 means there was one
     */
    if ((ret = nashParseSysfsDevno(sysfs_path, &devno)) <= 0)
        return ret;

    /* we can't just assign it to *dev, or gcc decides it's unused */
    tmp = calloc(1, sizeof (struct nash_block_dev));
    tmp->type = ADD;
    tmp->devno = devno;
    tmp->sysfs_path = strdup(sysfs_path);
    asprintf(&tmp->dev_path, "/dev/%s", strrchr(sysfs_path, '/')+1);
    while ((bang = strchr(tmp->dev_path, '!')))
        *bang = '/';
    *dev = tmp;
    return 1;
}

static inline int
nash_dev_tree_add_sysfs_dir_helper(nashContext *nc, char *path, int toplevel)
{
    struct nash_block_dev *bdev = NULL;
    int ret;
    int errnum;

    if (!strlimitcmp("/sys/block/ram", path) ||
            !strlimitcmp("/sys/block/loop", path))
        return 0;

    if (!toplevel) {
        ret = try_sysfs_dir(path, &bdev);
        if (ret > 0)
            nash_dev_tree_process_bdev(nc, bdev);
    }

    if (toplevel || ret >= 0) {
        DIR *dir;
        struct dirent *dent;

        if (!(dir = opendir(path)))
            return -1;

        while ((dent = readdir(dir))) {
            if (!strcmp(dent->d_name, ".") || !strcmp(dent->d_name, ".."))
                continue;

            size_t l = strlen(path) + 1 + strlen(dent->d_name) + 1;
            char subdir[l];
            sprintf(subdir, "%s/%s", path, dent->d_name);
            ret = nash_dev_tree_add_sysfs_dir_helper(nc, subdir, 0);
            if (ret < 0 && errno != ENOENT)
                break;
        }
        save_errno(errnum, closedir(dir));
    }
    return ret;
}

int
nash_dev_tree_add_sysfs_dir(nashContext *nc, char *path)
{
    nassert(path && path[0]);

    return nash_dev_tree_add_sysfs_dir_helper(nc, path, 1);
}

void nash_dev_tree_populate_from_sysfs(nashContext *nc)
{
    if (nash_dev_tree_add_sysfs_dir(nc, "/sys/block") < 0)
        return;

    if (nc->blktab) {
        int i;
        struct nash_dev_node *node;
        for (i = 0; !(node = nash_list_get(nc->devices->devs, i)); i++)
            nash_dev_tree_group_nodes(nc, node);
    }
}

void nash_dev_tree_free_ptr(nashContext *nc, struct nash_dev_tree **treep)
{
    struct nash_dev_tree *tree = *treep;

    if (!tree)
        return;

    if (tree->root)
        nash_dev_node_free_ptr(nc, &tree->root);

    if (tree->devs) {
        struct nash_dev_node *node;
        while ((node = nash_list_get(tree->devs, 0)))
            nash_dev_node_free_ptr(nc, &node);
    }
    if (tree->devs)
        nash_list_destroy(tree->devs);
    if (tree->fs_devs)
        nash_list_destroy(tree->fs_devs);

    memset(tree, '\0', sizeof (tree));
    free(tree);

    *treep = NULL;
}

struct nash_dev_tree *nash_dev_tree_alloc(nashContext *nc)
{
    struct nash_dev_tree *dev_tree;
    int errnum;

    struct blkent be = {
        .blk_name = "devtree", 
        .blk_type = "devtree",
        .blk_opts = "",
    };

    if (!(dev_tree = calloc(1, sizeof (*dev_tree))))
        return NULL;

    if (!(dev_tree->devs = nash_ptr_list_new())) {
err:
        save_errno(errnum, nash_dev_tree_free_ptr(nc, &dev_tree));
        return NULL;
    }

    if (!(dev_tree->fs_devs = nash_ptr_list_new()))
        goto err;

    if (!(dev_tree->root = nash_dev_node_from_blkent(nc, &be)))
        goto err;

    dev_tree->root->tree = dev_tree;

    return dev_tree;
}

/* 
 * Second: info about a device node
 */


static const char *nash_dev_type_names[] = {
    [DEV_TYPE_TREE] = "devtree",
    [DEV_TYPE_DM_ERROR] = "dm-error",
    [DEV_TYPE_DISK] = "disk",
    [DEV_TYPE_DM] = "dm",
    [DEV_TYPE_DM_MPATH] = "dm-mpath",
    [DEV_TYPE_PARTITION] = "part",
    [DEV_TYPE_LVM2_PV] = "lvm2-pv",
    [DEV_TYPE_LVM2_VG] = "lvm2-vg",
    [DEV_TYPE_LVM2_LV] = "lvm2-lv",
    [DEV_TYPE_DM_RAID] = "dmraid",
    [DEV_TYPE_MD_RAID] = "md",
    [DEV_TYPE_FS] = "fs",
    [DEV_TYPE_FLOPPY] = "floppy",
    [DEV_TYPE_NONE] = "invalid",
};

const char *nash_dev_type_name(int type)
{
    if (type < DEV_TYPE_TREE || type > DEV_TYPE_NONE)
        return NULL;
    return nash_dev_type_names[type];
}

void nash_dev_node_free_ptr(nashContext *nc, struct nash_dev_node **nodep)
{
    struct nash_dev_node *node = *nodep;
    struct nash_bdev_vitals *vitals;

    if (!node)
        return;

    xfree(node->name);

    if (node->vitals) {
        while ((vitals = nash_list_get(node->vitals, 0))) {
            nash_bdev_vitals_free(nc, vitals);
            nash_list_remove(node->vitals, 0);
        }
        nash_list_destroy(node->vitals);
    }

    nashBdevFreePtr(&node->bdev);
    
    nash_dev_tree_delete_node(nc, node);

    if (node->parents)
        nash_list_destroy(node->parents);
    if (node->parent_names)
        nash_list_destroy(node->parent_names);
    if (node->children)
        nash_list_destroy(node->children);
    if (node->child_names)
        nash_list_destroy(node->child_names);

    xfree(node);
    *nodep = NULL;
}

static struct nash_dev_node *nash_dev_node_alloc(nashContext *nc,
    char *name)
{
    struct nash_dev_node *node;
    int errnum;

    if (!(node = calloc(1, sizeof (*node))))
        return NULL;

    node->number = -1;
    if (name && !(node->name = strdup(name))) {
err:
        save_errno(errnum, nash_dev_node_free_ptr(nc, &node));
        return NULL;
    }
    if (!(node->vitals = nash_ptr_list_new()))
        goto err;

    return node;
}

int nash_dev_node_get_type_from_blkent(struct blkent *blkent)
{
    int i;
    for (i = DEV_TYPE_TREE; i < DEV_TYPE_NONE; i++)
        if (!strcmp(blkent->blk_type, nash_dev_type_name(i)))
            return i;
    return i;
}

static struct nash_dev_node *nash_dev_node_from_blkent(nashContext *nc,
        struct blkent *blkent)
{
    struct nash_dev_node *node;
    struct nash_bdev_vitals *vitals;
    int errnum;
    char *name = blkent ? blkent->blk_name : NULL;

    if (!(node = nash_dev_node_alloc(nc, name)))
        return NULL;

    node->tree = nc->blktab;

    node->flags = DEV_FLAG_ACCESS_OK | DEV_FLAG_MISSING;

    node->type = nash_dev_node_get_type_from_blkent(blkent);

    if (!(vitals = nash_bdev_vitals_from_blkent(nc, blkent))) {
err:
        save_errno(errnum, nash_dev_node_free_ptr(nc, &node));
        return NULL;
    }

    if (nash_list_add(node->vitals, vitals) < 0)
        goto err;

    if (!(node->parents = nash_ptr_list_new()))
        goto err;

    if (!(node->children = nash_ptr_list_new()))
        goto err;

    if (blkent) {
        char *start;

        errno = 0;
        if ((start = dupblkopt(blkent->blk_opts, "devs"))) {
            if (errno != 0)
                goto err;
            node->parent_names = nash_str_list_from_string(start, ",");
            save_errno(errnum, free(start));
            if (!node->parent_names)
                goto err;
        }
    }

    if (!node->parent_names && !(node->parent_names = nash_str_list_new()))
        goto err;
    if (!(node->child_names = nash_str_list_new()))
        goto err;

    return node;
}

static int devno_cmp(dev_t left, dev_t right)
{
    int x;

    if ((x = major(right) - major(left)))
        return x;
    x = minor(right);
    if (x == 0)
        return 0;
    return x - minor(left);
}

static int looks_like_partition(struct nash_block_dev *bdev) {
    char *dev_path = strrchr(bdev->dev_path, '/');

    if (!dev_path || !*dev_path || !*(++dev_path))
        return 0;
    for (; *dev_path && isalpha(*dev_path); dev_path++)
        ;
    if (!isdigit(*dev_path))
        return 0;
    for (; *dev_path; dev_path++) {
        if (!isdigit(*dev_path))
            return 0;
    }
    return 1;
}

static int looks_like_p_partition(struct nash_block_dev *bdev) {
    char *lastp = strrchr(bdev->dev_path, 'p');

    if (!lastp)
        return 0;
    while (*(++lastp)) {
        if (!isdigit(*lastp))
            return 0;
    }
    return 1;
}

static int sysfs_path_cmp(struct nash_dev_node *adult,
    struct nash_dev_node *child, int longest)
{
    int l,m;
    if (!adult->bdev || !adult->bdev->sysfs_path ||
            !child->bdev || !child->bdev->sysfs_path)
        return 0;
    l = strlen(adult->bdev->sysfs_path);
    m = strlen(child->bdev->sysfs_path);
    l = longest ? MAX(l,m) : MIN(l,m);
    if (!strncmp(adult->bdev->sysfs_path, child->bdev->sysfs_path, l))
        return 1;
    return 0;
}
static int is_sysfs_parent(struct nash_dev_node *adult,
        struct nash_dev_node *child)
{
    return sysfs_path_cmp(adult, child, 0);
}
static int is_sysfs_node(struct nash_dev_node *adult,
        struct nash_dev_node *child)
{
    return sysfs_path_cmp(adult, child, 1);
}

/* FIXME -ldmraid currently won't get us a list of formats.  That needs to
 * be fixed, and then this needs to query from it. */
static int looks_like_dmraid(struct nash_block_dev *bdev) {
    char *lasts = strrchr(bdev->dev_path, '/');
    static const char *prefixes[] = {
        "asr_", "hpt37x_", "hpt45x_", "isw_", "jmicron_", "lsi_",
        "nvidia_", "pdc_", "sil_", "via_", "ddf1_", ""
    };
    const char *prefix;

    if (!lasts || !*(++lasts))
        return 0;
    for (prefix = prefixes[0]; prefix[0] != '\0'; prefix++) {
        if (!strncmp(lasts, prefix, strlen(prefix)))
            return 1;
    }
    return 0;
}

/* XXX hahahahahahah this is terrible.  But there's no way to ask a dm
 * rule if it's lvm at this point... well, sortof.
 *
 * So, this *really* needs to go after all the other (better) checks...
 */
static int looks_like_lvm(struct nash_block_dev *bdev) {
    char *lasts = strrchr(bdev->dev_path, '/');

    if (!lasts || !(++lasts))
        return 0;
    if (strchr(lasts, '-'))
        return 1;
    return 0;
}

static int nash_dev_node_guess_type_from_bdev(nashContext *nc, 
        struct nash_dev_node *node)
{
    struct proc_dev_info const *info;
    int type;    
    char *name;
    dev_t devno;
    int m;

    devno = node->bdev->devno;
    m = -1;
    do {
        name = NULL;
        info = NULL;
        if ((m = getDevsFromProc(nc, m, S_IFBLK, &name, &devno, &info)) < 0)
            break;
        if (!info) {
            type = DEV_TYPE_NONE;
            if (looks_like_partition(node->bdev))
                type = DEV_TYPE_PARTITION;
        } else {
            type = info->type;
            if (info->flags & DEV_MAY_HAVE_PARTS &&
                    !(info->flags & DEV_USES_P_CONV)) {
                if (looks_like_partition(node->bdev))
                    type = DEV_TYPE_PARTITION;
            } else if (info->flags & DEV_USES_P_CONV) {
                if (type == DEV_TYPE_DM) {
                    char *dmtype;
                    if (!node->bdev || !node->bdev->dev_path)
                        return type;
                    dmtype = nashDmDevGetType(nc, node->bdev->devno);
                    if (dmtype && !strcmp(dmtype, "error"))
                        type = DEV_TYPE_DM_ERROR;
                    else if (looks_like_p_partition(node->bdev))
                        type = DEV_TYPE_PARTITION;
                    else if (looks_like_dmraid(node->bdev))
                        type = DEV_TYPE_DM_RAID;
                    else if (looks_like_lvm(node->bdev))
                        type = DEV_TYPE_LVM2_LV;
                    else
                        type = DEV_TYPE_DM_MPATH;
                } else {
                    if (looks_like_partition(node->bdev))
                        type = DEV_TYPE_PARTITION;
                }
            }
        }
        return type;
    } while (1);

    return DEV_TYPE_NONE;
}

static int nash_dev_node_add_vitals_from_bdev(
        nashContext *nc, struct nash_dev_node *node)
{
    struct nash_list *probed_vitals;
    struct nash_bdev_vitals *vitals;

#if 0
    fprintf(stderr, "guessing type for %s\n", node->bdev->sysfs_path);
#endif
    node->type = nash_dev_node_guess_type_from_bdev(nc, node);

#if 0
    fprintf(stderr, "probing vitals for %s\n", node->bdev->sysfs_path);
#endif
    if (!(probed_vitals = nash_vitals_probe(nc, node))) {
#if 0
        fprintf(stderr, "probed_vitals was null\n");
#endif
        return 0;
    }

    while ((vitals = nash_list_get(probed_vitals, 0))) {
#if 0
        fprintf(stderr, "found vitals %p for %s\n", vitals, node->bdev->sysfs_path);
#endif
        if (!nash_list_in(node->vitals, vitals)) {
#if 0
            fprintf(stderr, "vitals %p was not in list\n", vitals);
#endif
            nash_list_add(node->vitals, vitals);
            nash_list_remove(probed_vitals, 0);
        } else {
#if 0
            fprintf(stderr, "vitals %p was in list\n", vitals);
#endif
            nash_list_remove(probed_vitals, 0);
            nash_vitals_decref(nc, vitals);
        }
    }
    nash_list_destroy(probed_vitals);
    return 1;
}

static struct nash_dev_node *nash_dev_node_from_bdev(nashContext *nc,
        struct nash_block_dev *bdev)
{
    struct nash_dev_node *node;
    int errnum;

    smartmknod(bdev->dev_path, S_IFBLK | 0700, bdev->devno);
    if (!(node = nash_dev_node_alloc(nc, bdev->sysfs_path)))
        return NULL;

    node->tree = nc->devices;

    node->flags = DEV_FLAG_PROTECTED | DEV_FLAG_CHILDREN_MISSING;

    node->bdev = bdev;

    if (!(node->parents = nash_ptr_list_new()))
        goto err;
    if (!(node->parent_names = nash_str_list_new()))
        goto err;

    if (!(node->children = nash_ptr_list_new()))
        goto err;
    if (!(node->child_names = nash_ptr_list_new()))
        goto err;

#if 0
    fprintf(stderr, "adding vitals for %s from bdev\n", node->bdev->sysfs_path);
#endif
    nash_dev_node_add_vitals_from_bdev(nc, node);

    return node;
err:
    save_errno(errnum, nash_dev_node_free_ptr(nc, &node));
    return NULL;
}

/*
 * the heraldry concept here is that the closer it is to a bare disk,
 * the more adult it is.  adults provide resources (e.g. blocks) for
 * their children.
 *
 * That is to say, nothing has a disk as a child, and filesystems do not
 * have children.
 *
 * The table can be read as "a CHILD can reside on MIN to MAX ADULTs"
 * MAX of  0 means it's an invalid combination
 * MAX of -1 means unbounded
 *
 * child of type DEV_TYPE_NONE means end-of-list
 */
static const struct devtree_constraint {
    nash_dev_type child;
    nash_dev_type adult;
    int min_parents;
    int max_parents;
} devtree_constraints[] = {
    { DEV_TYPE_DISK, DEV_TYPE_TREE, 0, -1 },
    { DEV_TYPE_DM_MPATH, DEV_TYPE_DISK, 2, -1 },
    { DEV_TYPE_DM_RAID, DEV_TYPE_DISK, 1, -1 },
    { DEV_TYPE_PARTITION, DEV_TYPE_DISK, 1, 1 },
    { DEV_TYPE_PARTITION, DEV_TYPE_DM_MPATH, 1, 1 },
    { DEV_TYPE_PARTITION, DEV_TYPE_DM_RAID, 1, 1 },
#if 0 || defined(NASH_SUPPORT_ABSURD_DEVICE_TOPOLOGY)
    { DEV_TYPE_PARTITION, DEV_TYPE_MD_RAID, 1, 1 },
#endif
    { DEV_TYPE_MD_RAID, DEV_TYPE_PARTITION, 1, -1 },
#if 0 || defined(NASH_SUPPORT_ABSURD_DEVICE_TOPOLOGY)
    { DEV_TYPE_LVM2_PV, DEV_TYPE_DISK, 1, 1 },
    { DEV_TYPE_LVM2_PV, DEV_TYPE_DM_MPATH, 1, 1 },
    { DEV_TYPE_LVM2_PV, DEV_TYPE_DM_RAID, 1, 1 },
    { DEV_TYPE_LVM2_PV, DEV_TYPE_MD_RAID, 1, 1 },
#endif
    { DEV_TYPE_LVM2_PV, DEV_TYPE_PARTITION, 1, 1 },
    { DEV_TYPE_LVM2_PV, DEV_TYPE_MD_RAID, 1, 1 },
    { DEV_TYPE_LVM2_VG, DEV_TYPE_LVM2_PV, 1, -1 },
    { DEV_TYPE_LVM2_LV, DEV_TYPE_LVM2_VG, 1, 1 },
#if 0 || defined(NASH_SUPPORT_ABSURD_DEVICE_TOPOLOGY)
    { DEV_TYPE_FS, DEV_TYPE_DISK, 1, 1 },
    { DEV_TYPE_FS, DEV_TYPE_DM_MPATH, 1, 1 },
    { DEV_TYPE_FS, DEV_TYPE_DM_RAID, 1, 1 },
#endif
    { DEV_TYPE_FS, DEV_TYPE_LVM2_LV, 1, 1 },
    { DEV_TYPE_FS, DEV_TYPE_PARTITION, 1, 1 },
    { DEV_TYPE_FS, DEV_TYPE_MD_RAID, 1, 1 },
    { DEV_TYPE_NONE, DEV_TYPE_NONE, 0, 0 }
};

static int is_sysfs_slave(struct nash_dev_node *adult,
    struct nash_dev_node *child)
{
    DIR *dir;
    struct dirent *entry;
    struct stat sb;
    int rc = 0;
    char slaves[strlen(child->bdev->sysfs_path) + strlen("/slaves") + 1];

    strcpy(slaves, child->bdev->sysfs_path);
    strcat(slaves, "/slaves");

    if (!(dir = opendir(slaves)))
        return 0;

    while ((entry = readdir(dir))) {
        char slave[strlen(slaves) + 1 + strlen(entry->d_name) + 1];
        char devpath[sizeof(slave) + strlen("/dev") + 1];
        char devbuf[32] = "";
        int maj, min;
        FILE *df;

        strcpy(slave, slaves); 
        strcat(slave, "/");
        strcat(slave, entry->d_name);
        strcpy(devpath, slave);
        strcat(devpath, "/dev");

        lstat(slave, &sb);
        if (!S_ISLNK(sb.st_mode))
            continue;

        if (!(df = fopen(devpath, "r")))
            continue;

        if (!fgets(devbuf, 31, df)) {
            fclose(df);
            continue;
        }
        fclose(df);

        if (sscanf(devbuf, "%d:%d", &maj, &min) != 2)
            continue;

        if (makedev(maj,min) == adult->bdev->devno) {
            rc = 1;
            break;
        }
    }
    closedir(dir);
    return rc;
}

static int nash_dev_node_viable_parent(nashContext *nc,
    struct nash_dev_node *adult, struct nash_dev_node *child)
{
    int i;
    struct nash_dev_node *parent;
    int num_named_parents = 0;
    int num_found_parents = 0;

    if (child == adult)
        return 0;

    num_named_parents = -1;
    if (child->tree == nc->devices) {
        if (!child->bdev || !child->bdev->sysfs_path ||
                !adult->bdev || !adult->bdev->sysfs_path)
            return 0;
        if (!strncmp(adult->bdev->sysfs_path, child->bdev->sysfs_path, 
                strlen(adult->bdev->sysfs_path)))
            return 1;
        if (is_sysfs_slave(adult, child))
            return 1;
        return 0;
    } else {
        num_named_parents = nash_list_len(child->parent_names);
    }
    num_found_parents = nash_list_len(child->parents);

    if (num_named_parents == 0 || num_named_parents == num_found_parents)
        return 0;

    parent = nash_list_get(child->parents, 0);

    for (i = 0; devtree_constraints[i].child != DEV_TYPE_NONE; i++) {
        const struct devtree_constraint *con = &devtree_constraints[i];

        if (con->child != child->type || con->adult != adult->type)
            continue;
        if (con->max_parents == 0)
            continue;
        if (num_found_parents > 0 && parent) {
            /* if there are already parents, but they're not this type,
             * then this can't be it... */
            if (con->adult != parent->type)
                continue;

            if (con->max_parents != -1 && con->max_parents <= num_found_parents)
                continue;
        }

        if (adult->tree == nc->devices) {
            if (is_sysfs_parent(adult, child))
                return 1;
        } else {
            if (nash_list_in(child->parent_names, adult->name))
                return 1;
        }
    }
    return 0;
}

/* purpose: remove child from its parents */
static void nash_dev_node_emancipate_from_parents(nashContext *nc,
    struct nash_dev_node *child)
{
    struct nash_dev_node *adult;
    int pos;

    if (!child->parents)
        return;
    while ((adult = nash_list_get(child->parents, 0))) {
        pos = nash_list_search(adult->children, child);
        nash_list_remove(adult->children, pos);
        nash_list_remove(child->parents, 0);
        if (adult->tree == nc->devices)
            adult->flags &= ~DEV_FLAG_CHILDREN_MISSING;
    }
    child->flags &= ~(DEV_FLAG_IN_TREE);
}

/* purpose: remove adult's children from adult */
static void nash_dev_node_surrender_children(nashContext *nc, 
    struct nash_dev_node *adult)
{
    struct nash_dev_node *child;
    int i;

    if (!adult->children)
        return;
    while ((child = nash_list_get(adult->children, 0))) {
        i = nash_list_search(child->parents, adult);
        nash_list_remove(child->parents, i);
        nash_list_remove(adult->children, 0);
        child->flags &= ~DEV_FLAG_IN_TREE;
        if (child->tree == nc->devices)
            child->flags &= ~DEV_FLAG_PARENTS_MISSING;
    }
}

char *nash_dev_node_get_dm_name(nashContext *nc, char *name)
{
    struct nash_dev_node *node;
    char *dmname = NULL;
    char *tmp = NULL;
    int i = -1;
    int blktab = 0;
    
    if (name[0] == '@') {
        name += 1;
        blktab = 1;
    }
    node = nash_dev_tree_find_device(nc, name, blktab);
    while (nash_dev_node_vitals_get(nc, node, "device_node", ++i, &tmp) >= 0) {
        if (tmp)
            break;
    }
    if (!tmp)
        tmp = name;
    dmname = strrchr(tmp, '/');
    if (dmname && dmname[0] == '/')
        dmname++;
    if (!dmname || !dmname[0])
        dmname = tmp;
    return dmname;
}

/*
 * Third: loading the blktab from disk.
 */

static int nash_load_blkents(nashContext *nc, const char *path)
{
    FILE *btfile = NULL;
    struct blkent *blkent = NULL;
    int i = 0;

    if (!(btfile = setblkent(path, "r")))
        return 0;

    while ((blkent = getblkent(btfile)) != NULL) {
        nash_dev_tree_add_blkent(nc, blkent);
        i++;
    }

    endblkent(btfile);
    return i;
}

int nash_load_blktab(nashContext *nc)
{
    int i = 0, ret = 0;

    if (nc->blktab)
        return 0;

    if ((nc->blktab = nash_dev_tree_alloc(nc)) < 0)
        return -1;

    ret = nash_load_blkents(nc, "/etc/blktab");

    if (i > 0 && nc->devices) {
        int i;
        struct nash_dev_node *node;
        for (i = 0; !(node = nash_list_get(nc->devices->devs, i)); i++)
            nash_dev_tree_group_nodes(nc, node);
    }

    return ret;
}

/*
 * fourth section:
 * seeing if we have complete info for a device, and if so, building it.
 */
/*
 * return codes are as follows:
 * 1 == can build (or at least we think so ;)
 * 0 == can partially build (e.g. for md, dmraid, or mpath)
 * -1 == no way at all.
 */
/* XXX this really only checks if all the DEV_TYPE_DISK devices are present */
int nash_dev_tree_check_node_complete(nashContext *nc,
        struct nash_dev_node *child)
{
    int rc = 1;
    struct nash_dev_node *adult;
    int i, l0, l1;

    if (child->flags & DEV_FLAG_COMPLETE)
        return 1;

    if (child->bdev)
        return -1;

    switch (child->type) {
        case DEV_TYPE_TREE:
        case DEV_TYPE_NONE:
            return -1;
        case DEV_TYPE_DISK:
            if (child->mate == NULL)
                return -1;
            child->flags |= DEV_FLAG_COMPLETE;
            child->mate->flags |= DEV_FLAG_COMPLETE;
            return 1;
        case DEV_TYPE_MD_RAID:
            /* we have to handle md raid differently for now, because "md0"
               is always present */
#if 0
            if (nash_list_len(child->parents) == 0)
                return -1;
#endif
            /* fall through */
        case DEV_TYPE_DM_MPATH:
        case DEV_TYPE_DM_RAID:
        case DEV_TYPE_PARTITION:
        case DEV_TYPE_LVM2_VG:
        case DEV_TYPE_LVM2_LV:
        case DEV_TYPE_FS:
        default:
            i = -1;
            while (rc >= 0 && (adult = nash_list_get(child->parents, ++i))) {
                int ret;
                ret = nash_dev_tree_check_node_complete(nc, adult);
                if (rc > ret)
                    rc = ret;
                if (rc < 0)
                    break;
            }
            if (rc >= 0) {
                l0 = nash_list_len(child->parent_names);
                l1 = nash_list_len(child->parents);
                if (l0 == l1) {
                    rc = MIN(rc, 1);
                } else {
                    if (l0 > 0)
                        rc = 0;
                    else
                        rc = -1;
                }
            }
            if (rc == 1)
                child->flags |= DEV_FLAG_COMPLETE;
            return rc;
    }
    return -1;
}

static dev_t get_devno(char *path)
{
    struct stat sb;
    dev_t ret = makedev(-1,-1);

    if (stat(path, &sb) < 0)
        return ret;
    if (!S_ISBLK(sb.st_mode))
        return ret;
    return sb.st_rdev;
}

/*
 * find a tree node given the label, uuid, or path
 */
struct nash_dev_node *nash_dev_tree_find_device(nashContext *nc, char *device,
        int blktab)
{
    int i;
    struct nash_dev_node *child, *adult;
    char namebuf[1024] = {'\0'};
    enum {
        LABEL,
        UUID,
        PATH,
    } type = PATH;

    i = -1;
    adult = NULL;

    if (blktab) {
        while (!adult && (child = nash_list_get(nc->blktab->devs, ++i))) {
            if (!strcmp(child->name, device)) {
                adult = child;
                break;
            }
        }
    } else {
        char *target = device;

        if (!strncasecmp(device, "label=", 6)) {
            type = LABEL;
            target += 6;
        } else if (!strncasecmp(device, "uuid=", 5)) {
            type = UUID;
            target += 5;
        }

        if (type != PATH) {
            while (!adult &&
                    (child = nash_list_get(nc->devices->fs_devs, ++i))) {
                if (nash_dev_node_vitals_has(nc, child,
                        type == LABEL ? "label" : "unique_id", target))
                    adult = child;
            }
            while (!adult &&
                    (child = nash_list_get(nc->devices->devs, ++i))) {
                if (nash_dev_node_vitals_has(nc, child,
                        type == LABEL ? "label" : "unique_id", target))
                    adult = child;
            }
        } else {
            dev_t devno = get_devno(device);
            dev_t ndev = makedev(-1,-1);

            while (!adult && (child = nash_list_get(nc->devices->devs, ++i))) {
                char *slash, *lvname;

                if (devno != ndev && devno == child->bdev->devno) {
                    adult = child;
                    break;
                } else if ((child->bdev && child->bdev->dev_path &&
                            !strcmp(child->bdev->dev_path, device)) ||
                        (child->mate && child->mate->bdev &&
                         child->mate->bdev->dev_path &&
                         !strcmp(child->mate->bdev->dev_path, device))) {
                    adult = child;
                    break;
                } else if ((child->bdev && child->bdev->sysfs_path &&
                            !strcmp(child->bdev->sysfs_path, device)) ||
                        (child->mate && child->mate->bdev &&
                         child->mate->bdev->sysfs_path &&
                         !strcmp(child->mate->bdev->sysfs_path, device))) {
                    adult = child;
                    break;
                }

                if (!adult && (lvname = strrchr(device, '/'))) {
                    slash = strchr(device, '/');
                    strncpy(namebuf, slash, MIN(lvname-slash, 1023));
                    lvname++;
                    if (child->type == DEV_TYPE_LVM2_LV &&
                            !strcmp(child->name, lvname)) {
                        child = nash_list_get(child->parents, 0);
                        if (child->type == DEV_TYPE_LVM2_VG &&
                                !strcmp(child->name, namebuf))
                            adult = child;
                    }
                }
            }
        }
    }

    return adult;
}

static int nash_dev_node_setup_parents(nashContext *nc,
    struct nash_dev_node *child)
{
    int rc = 1;

    if (child->type == DEV_TYPE_DISK) {
        if (child->mate) {
            child->flags |= DEV_FLAG_COMPLETE;
            child->mate->flags |= DEV_FLAG_COMPLETE;
            return 1;
        }
        return 0;
    } else {
        struct nash_dev_node *parent;
        int i = -1;
        while ((parent = nash_list_get(child->parents, ++i))) {
            int ret = nash_dev_node_setup_parents(nc, parent);
            if (ret < rc)
                rc = ret;
        }
    }
    return rc;
}

int nash_dev_tree_setup_device(nashContext *nc, char *device)
{
    struct nash_dev_node *node = NULL;
    int blktab = 0;

    if (device[0] == '@') {
        device += 1;
        blktab = 1;
    }

    node = nash_dev_tree_find_device(nc, device, blktab);
    if (!node) {
        nashLogger(nc, NASH_ERROR, "Could not find device %s%s\n",
                blktab ? "@" : "", device);
        return -1;
    }
    int rc;
    if ((rc = nash_dev_node_setup_parents(nc, node)) < 0)
        nashLogger(nc, NASH_ERROR, "Could not setup parents of device %s\n",
                node->name);
    return rc;
}

char *nash_dev_tree_replace_dm_params(nashContext *nc, long long start,
        long long length, char *params)
{
    struct nash_dev_node *node;
    char *ret = NULL, *tmp = NULL;
    char *prev = params, *atsign;

    while ((atsign = strchr(prev, '@'))) {
        int l = strcspn(atsign, " \t\n");
        char *dmname = NULL;
        dev_t devno;
        int devmaj, devmin;
        char c;

        c = atsign[l];
        atsign[0] = '\0';
        atsign[l] = '\0';

        node = nash_dev_tree_find_device(nc, atsign+1, 1);

        if (node->mate) {
            devno = node->mate->bdev->devno;
        } else {
            dmname = node->name;
            nashLogger(nc, NASH_ERROR, "Could not find device %s\n", dmname);
            nashDmCreate(nc, dmname, NULL, start, (uint64_t)-1, "error", "");
            node->flags |= DEV_FLAG_COMPLETE;

            devno = nashDmGetDev(dmname);
        }
        devmaj = major(devno);
        devmin = minor(devno);

        asprintf(&tmp, "%s%s%d:%d", ret ? ret : "", prev, devmaj, devmin);
        atsign[0] = '@';
        
        if (ret)
            free(ret);
        ret = tmp;
        atsign[l] = c;
        prev = atsign + l;
    }
    if (ret) {
        if (prev[0] != '\0') {
            asprintf(&tmp, "%s%s", ret, prev);
            free(ret);
            ret = tmp;
        }
    } else {
        ret = strdup(prev);
    }
    return ret;
}

static int mpathn = 0, dmraidn = 0, lvn = 0, fsn = 0;

static void nash_dev_node_set_number(struct nash_dev_node *node,
        int *n)
{
    if (node->number == -1)
        return;
    node->number = (*n)++;
}

static void print_node_name(struct nash_dev_node *node)
{
    char *c;
    switch (node->type) {
        case DEV_TYPE_DISK:
            c = strrchr(node->bdev->sysfs_path, '/') + 1;
            printf("%s", c);
            break;
        case DEV_TYPE_DM_MPATH:
            printf("mpath%d", node->number);
            break;
        case DEV_TYPE_DM_RAID:
            printf("dmraid%d", node->number);
            break;
        case DEV_TYPE_PARTITION:
            c = strrchr(node->bdev->sysfs_path, '/') + 1;
            printf("%s", c);
            break;
        case DEV_TYPE_LVM2_PV:
            printf("pv%d", node->number);
            break;
        case DEV_TYPE_LVM2_VG:
            printf("vg%d", node->number);
            break;
        case DEV_TYPE_LVM2_LV:
            printf("lv%d", node->number);
            break;
        case DEV_TYPE_MD_RAID:
            c = strrchr(node->bdev->sysfs_path, '/') + 1;
            printf("%s", c);
            break;
        case DEV_TYPE_FS:
            printf("fs%d", node->number);
        default:
            break;
    }
}

void print_parents(struct nash_dev_node *child)
{
    struct nash_list_iter *iter = NULL;
    struct nash_dev_node *parent;
    int len;

    len = nash_list_len(child->parents);
    if (len) {
        printf("devs=");
        while ((iter = nash_list_iter_next(child->parents, iter))) {
            parent = nash_list_iter_data(iter);
            print_node_name(parent);
            if (nash_list_iter_pos(iter) > len - 1)
                printf(",");
        }
    }
}

static void nash_dev_node_print_blktab(nashContext *nc,
        struct nash_dev_node *child, int parents)
{
    struct nash_list_iter *iter = NULL;
    struct nash_dev_node *parent;

    if (parents) {
        while ((iter = nash_list_iter_next(child->parents, iter))) {
            parent = nash_list_iter_data(iter);

            nash_dev_node_print_blktab(nc, parent, parents);
        }
    }

    switch (child->type) {
        case DEV_TYPE_DISK:
            print_node_name(child);
            printf("\tdisk\t");
            printf("device_node=%s;", child->bdev->dev_path);
            print_parents(child);
            while ((iter = nash_list_iter_next(child->vitals, iter))) {
                struct nash_bdev_vitals *vitals = nash_list_iter_data(iter);
                if (vitals->type == DEV_TYPE_DISK &&
                        vitals->unique_id && vitals->disk_vendor &&
                        vitals->disk_model && vitals->disk_probe_module &&
                        vitals->disk_probe_name) {
                    printf(";model=%s;unique_id=%s;vendor=%s",
                        vitals->disk_model, vitals->unique_id,
                        vitals->disk_vendor);
                    printf(";module=%s;probe=%s", vitals->disk_probe_module,
                        vitals->disk_probe_name);
                    nash_list_iter_end(iter);
                    iter = NULL;
                    break;
                }
            }
            printf("\n");
            break;
        case DEV_TYPE_DM_MPATH:
            nash_dev_node_set_number(child, &mpathn);
            print_node_name(child);
            printf("\n");
            break;
        case DEV_TYPE_DM_RAID:
            nash_dev_node_set_number(child, &dmraidn);
            print_node_name(child);
            printf("\n");
            break;
        case DEV_TYPE_PARTITION:
            print_node_name(child);
            printf("\tpart\t");
            print_parents(child);
            printf("\n");
            break;
        case DEV_TYPE_LVM2_PV:
            /* this is handled by mkinitrd */
            break;
        case DEV_TYPE_LVM2_VG:
            /* this is handled by mkinitrd */
            break;
        case DEV_TYPE_LVM2_LV:
            nash_dev_node_set_number(child, &lvn);
            print_node_name(child);
            printf("\tlvm2-lv\t");
            printf("device_node=%s", child->bdev->dev_path);
            printf("\n");
            break;
        case DEV_TYPE_MD_RAID:
            print_node_name(child);
            printf("\n");
            break;
        case DEV_TYPE_FS:
            nash_dev_node_set_number(child, &fsn);
            print_node_name(child);
            printf("\n");
            break;
        default:
            break;
    }
}

void nash_dev_tree_print_blktab(nashContext *nc, char *devname)
{
    struct nash_dev_node *node = nash_dev_tree_find_device(nc, devname, 0);
    struct nash_list_iter *iter = NULL;

    if (!node)
        return;

    nash_dev_node_print_blktab(nc, node, 1);
    
    while ((iter = nash_list_iter_next(node->children, iter))) {
        struct nash_dev_node *child = nash_list_iter_data(iter);

        nash_dev_node_print_blktab(nc, child, 0);
    }
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
